#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define LINELEN 128

typedef struct
{
   char hint[LINELEN];
   char tip[LINELEN];
   char full[LINELEN];
} answer_type;

/* initialise global variables */
char **kwestions;
char **words;
FILE *filein;
answer_type *answers;
char name[LINELEN];
char author[LINELEN];
char info[LINELEN];
int num_questions;
int num_words;
char nullstr[10]="\0";
char stupid[LINELEN]="\0";
char infile[LINELEN];
int list_info=1;
int list_full=0;
int list_tip=0;
int list_hint=1;
int list_all=0;
int show_kwes=1;
int kwestion=0;

#define instr(x,y) (int) (strchr(x,y)-x)
#define streq(x,y) !strcmp(x,y)

char *left(char *string,int num)
{
   char temp[LINELEN]="\0";
   if (num>strlen(string) || num<=0) return string;
   strncat(temp,string,num);
   return temp;
}

char *rest(char *string,int num)
{
   if (num>strlen(string) || num<=0) return nullstr;
   return string+num;
}

void quit(void)
{
   if (filein!=NULL) fclose(filein);
   exit(0);
}

void *memalloc(int size)
{
   void *memory;

   memory=malloc(size);
   if (memory==NULL)
   {
      printf("Fatal Error: Ran out of memory...\n");
      quit();
   }
   return memory;
}

char *removes(char *string)
{
   char *temp=string;
   strcpy(temp,string);
   while (*temp==' ' && temp<strlen(string)+string) temp++;
   if (temp>=strlen(string)+string) temp=nullstr;
   return temp;
}

char *getanswer(char *string)
{
   char first[LINELEN],answer[LINELEN];
   int i;
   strcpy(answer,"\0");
   while(strlen(string)>0)
   {
      i=instr(string,' ');
      strcpy(first,left(string,i));
      strcpy(string,removes(rest(string,i+1)));
      if (atoi(first)>num_words) {printf("Error in parsing question\n"); quit();}
      strcat(answer,words[atoi(first)]);
      strcat(answer," ");
   }
   return answer;
}

void parsefile(void)
{
   char string[LINELEN]; /*quick way to malloc 128 bytes :) */
   char first[LINELEN];
   int i,j,cur_ques=0;

   while (!feof(filein))
   {
      fscanf(filein,"%[^\n]\n",string);
      switch (string[0])
      {
         case '#' :
            strcpy(string,rest(string,1));

            i=instr(string,' ');
            strcpy(first,left(string,i));
            /* I'm too lazy to do a hash table... :) */
            if (streq(first,"name"))
            {
               strcpy(name,rest(string,i+1));
               goto out;          /* argh :) */
            }

            if (streq(first,"author"))
            {
               strcpy(author,rest(string,i+1));
               goto out;          /* argh :) */
            }

            if (streq(first,"info"))
            {
               strcpy(info,rest(string,i+1));
               goto out;          /* argh :) */
            }

            if (streq(first,"questions"))
            {
               num_questions=atoi(rest(string,i+1));
               kwestions=memalloc(num_questions*sizeof(char *));
               for (j=0; j<num_questions; j++)
                  kwestions[j]=memalloc(LINELEN);
               answers=memalloc(num_questions*sizeof(answer_type));
               goto out;          /* argh :) */
            }

            if (streq(first,"words"))
            {
               num_words=atoi(rest(string,i+1));
               words=memalloc((num_words+1)*sizeof(char *));
               for (j=0; j<=num_words; j++)
                  words[j]=memalloc(LINELEN);
               goto out;          /* argh :) */
            }

            printf("Unknown directive: %s; ignoring\n",first);

out:        break;

         case '?' :
            if (kwestions==NULL) { printf("No #questions directive\n");quit(); }
            strcpy(string,rest(string,instr(string,' ')));
            strcpy(string,removes(string));
            strcpy(kwestions[cur_ques],string);
            fscanf(filein,"%[^\n]\n",answers[cur_ques].hint);
            fscanf(filein,"%[^\n]\n",answers[cur_ques].tip);
            fscanf(filein,"%[^\n]\n",answers[cur_ques].full);
            cur_ques++;
            break;

         default :
            if (words==NULL) { printf("No #words directive\n");quit(); }
	    strcpy(first,left(string,instr(string,' ')));
	    j=atoi(first);
	    strcpy(string,removes(rest(string,instr(string,' '))));
	    if (j>num_words) { printf("Number more than amount of words\n");quit();}
	    strcpy(words[j],string);
      }
   }
}

void syntax(char a[])
{
   printf("Syntax: *%s [flags] <hintfile> <question number>\n",a);
   printf("Flags are:\n");
   printf("\t-f         - list full solution to problem (long)\n");
   printf("\t-h|-?      - this message\n");
   printf("\t-i         - list hint to problem\n");
   printf("\t-l         - list all questions\n");
   printf("\t-k         - list question\n");
   printf("\t-t         - list tip to problem\n");
   printf("\t+d         - don't list hint file details\n");
   printf("\t+i         - don't list hint to solution\n");
   printf("\t+k         - don't show question\n");
   exit(0);
}

void parse_arguments(argc,argv)
int argc;
char *argv[];
{
   int carg=1,c=0,carg2;

   infile[0]='\0';

   while (carg<argc)
   {
      c=0;
      switch (argv[carg][0])
      {
         case '-' :
            carg2=carg;
            while (c<strlen(argv[carg])-1)
            {
               switch(argv[carg][++c])
               {
                  case 'h' :
                  case '?' :
                     syntax(argv[0]);

                  case 'l' :
                     list_all=1;
                     show_kwes=0;
                     list_hint=0;
                     break;

                  case 'f' :
                     list_full=1;
                     break;

                  case 'i' :
                     list_hint=1;
                     break;

                  case 'k' :
                     show_kwes=1;
                     break;

                  case 'q' :
                     printf("You've found the quote switch:\n");
                     printf("\t\"Do you believe in forever?\n");
                     printf("\t I don't even believe in tomorrow\n");
                     printf("\t The only thing that last forever are memories and sorrow\n");
                     printf("\t Out of sight out of mind the motto of betrayal\"\n");
                     printf("- from 'Unsuccessfully Coping with the Natural Beauty of Infidelity'\n  by Peter Steele\n");
                     exit(0);

                  case '-' :
                     strcpy(infile,argv[++carg2]);
                     break;

                  case 't' :
                     list_tip=1;
                     break;

                  default :
                     printf("Invalid flag: -%c\n",argv[carg][c]);
                     syntax(argv[0]);
               }
            }
            carg=carg2+1;
            break;

         case '+' :
            carg2=carg;
            while (c<strlen(argv[carg])-1)
            {
               switch(argv[carg][++c])
               {
                  case 'd' :
                     list_info=0;
                     break;

                  case 'i' :
                     list_hint=0;
                     break;

                  case 'k' :
                     show_kwes=0;
                     break;

                  default :
                     printf("Invalid flag: +%c\n",argv[carg][c]);
                     syntax(argv[0]);
                     break;
               }
            }
            carg=carg2+1;
            break;

         default :
            if (argc!=carg)
            {
               if (infile[0]=='\0') strcpy(infile,argv[carg++]);
               else
               {
                  if (kwestion==-1 || kwestion==0) kwestion=atoi(argv[carg++]);
                  else
                  {
                  printf("Invalid argument: %s\n",argv[carg]);
                  syntax(argv[0]);
                  }
               }
            }
      }
   }

   if (infile[0]=='\0')
   {
      printf("No input file specified\n");
      syntax(argv[0]);
   }

}

int main(int argc,char **argv)
{
   int i;
   parse_arguments(argc,argv);

   if ((filein=fopen(infile,"r"))==NULL)
   {
      printf("Cannot open read file...\n");
      quit();
   }

   parsefile();

   if (list_info)
   {
      printf("Game:        %s\n",name);
      printf("Hints by:    %s\n",author);
      printf("Other info:  %s\n\n",info);
   }

   if (kwestion>=num_questions)
   {
      printf("Too high a value of question\n");
      quit();
   }

   if (show_kwes) printf("Question %d:\n%s\n",kwestion,kwestions[kwestion]);

   /* don't ask me why I need the stupid var - but the string gets
      corrupted if I don't have it... */

   if (list_hint)
   {
      strcpy(stupid,getanswer(answers[kwestion].hint));
      printf("hint:        %s\n",stupid);
   }
   if (list_tip)
   {
      strcpy(stupid,getanswer(answers[kwestion].tip));
      printf("tip:         %s\n",stupid);
   }
   if (list_full)
   {
      strcpy(stupid,getanswer(answers[kwestion].full));
      printf("full:        %s\n",stupid);
   }

   if (list_all)
   {
      for (i=0;i<num_questions;i++)
         printf("%d \t: %s\n",i,kwestions[i]);
   }

   fclose(filein);
   return 1;
}

